* ͸
*  
*  Procedure:      DRAWMAP
*  
*  Author:         Simon Collingridge
*  Date:           10/5/94
*  
*  DESCRIPTION:
*  
*      This function is used to output the details of all the Table
*  relations that are set at the time of calling it. Initially it creates
*  a cursor which stores the details for each relation that is set. Once
*  this has been done and one or more relations have been found, this 
*  function then sets up the required variables and arrays to create the
*  relational diagram and associated information to be output into a
*  text file, before it calls the MAPRELS function.
*  
*          Ŀ
*           NOTE:     MAPRELS is a recursive function making  
*                     calls on itself. It is fully commented. 
*          
*  
*  
*      Once MAPRELS has complete execution, the results are output to a
*  file, the name of which is stored into a public memory variable named
*  m.cMapText. Until written to file, all the information is stored in a
*  private mempry variable called m.cMap. The reason for this is so output
*  can be easily re-directed to where it may be need (e.g. a memo field 
*  or screen).
*  
*      This procedure refers to relationships as being of a certain
*  level number. A top level parent (i.e. a Parent Table that is not
*  itself a child of another table) if considered to be level 1. Each 
*  subsequent level is numbered accordingly:
*  
*  e.g.:     Table1
*             Table2
*                Table3
*                   Table4
*                Table5
*                    Table6
*             Table7
*  
*      Table Levels
*  
*      Table 1    -    Level 1
*      Table 2    -    Level 2
*      Table 3    -    Level 3
*      Table 4    -    Level 4
*      Table 5    -    Level 3
*      Table 6    -    Level 4
*      Table 7    -    Level 2
*  
*  
*  Inputs:    None
*  
*  Outputs:
*  
*      Text file of the relations diagram. The file name is stored in 
*      PUBLIC VARIABLE m.cMapText. Output can be easily re-directed by 
*      changing the destination of the memory variable m.cMap to wherever
*      it is needed. The file name of the text file is generated from
*      a combination of sys(3) and sys(2015) and is given a .MAP extention.
*  
*      The format of the output is as follows:
*      
*      Ŀ
*                                                                      
*                      Table Relationships                             
*                                                                      
*                                                                      
*        Relation Field  Table Order  Diagram                        
*          
*                                                                    
*                                                                    
*                       <Index tag 1> Table1                         
*        <Field name 1> <Index tag 2>  Table2                     
*        <Field name 2> <Index tag 3>     Table3                 
*        <Field name 3> <Index tag 4>        Table4             
*        <Field name 4> <Index tag 5>     Table5                 
*        <Field name 5> <Index tag 6>         Table6             
*        <Field name 6> <Index tag 7>  Table7                     
*                                                                      
*      
*      
*      Where:
*      
*      Relation Field:
*      
*      The relation field is the name of the field used by the 
*      parent of the table opposite (in the diagram) to relate
*      INTO it. (i.e. :
*                      <Field name 1> is the field Table 1 uses to relate
*                                     INTO Table 2
*                      <Field name 2> is the field Table 2 uses to relate
*                                     INTO Table 3
*                      <Field name 5> is the field Table 2 uses to relate
*                                     INTO Table 5
*                
*      Table Order:
*      
*      The Table order is the order of the Table opposite in the diagram
*      i.e. 
*            <Index tag 1> is the index that Table 1 is ordered by.
*            <Index tag 6> is the index that Table 6 is ordered by.
*            
*            
*  Called By:    SAVE_IT
*  
*  Calls:        BUILDMAP
*  
*  Tables Used:  All those that are related and already open.
*  
*  Calling
*     Sequence:  DO DRAWMAP
*  
*  	
*  	
*  MODIFICATIONS:
*  	
*  	
*  	23 May 1994		Simon Collingridge
*  	
*  	The following modifications have been made:
*  	
*  	1.	The PRIVATE variable m.nMaxNest was added to store the maximum
*  		number of nested relations that can be mapped. See the 
*  		MODIFICATIONS section of the BUILDMAP procedure for details.
*  	
*  	2.	The first line of the DO WHILE loop that finds all the 
*  		individual relatons was changed from:
*  	
*  			do while target(m.nTgt, m.nArea) <> ""
*  		to:
*  	
*  			do while !empty(target(m.nTgt, m.nArea))
*  	
*  		This was done as a bugfix due to the way that the SET EXACT
*  		setting effected the result of the original comparrison.
*  	
*  	3.	Added a more descriptive meesage for the event that no 
*  		relations are mapped by DRAWPMAP.
*  	
*  	
*  	
* ;

public  m.cMapText

private m.nArea, m.nTgtf, m.pnFile, m.nCurRec, m.nCurrArea, m.nTopLevel, ;
		m.nWorkAreas, m.nMaxNest, m.cCrlf, m.cMap, m.cTalkSet, ;
		m.cCntStore, m.cStoreFrom, m.cCursName, m.lDoneOnce, m.lNotFound, ;
		na0LevTot, na0LevCnt, ca0LevPad, ca0Parent

* ͸
*                  MEMORY VARIABLE DEFINITIONS AND USE
*  
*  All variables use the following naming standard based on Hungarian
*  Notation:
*  
*  
*     Prefix    Data Type
*    
*             
*     m.c       Character
*     m.n       Numeric
*     m.l       Logical
*     m.d       Date
*               
*     ca#       Character Array
*     na#       Numeric Array
*     la#       Logical Array
*     da#       Date Array
*     ma#       Mixed Data Type Array
*               
*               # is the total number of elements (ALEN()) in the array, 
*                 zero "0" indicates a variable number of elements
*               
*     m.pn      Numeric File Pointer
*               
*     xt        Index Tag Name
*               
*  
*  
*  where m.cVarName would indicate a character data type
*         
*  
*  
*  PUBLIC variables:
*  
*    Name          Type  Use
*  
*    m.cMapText    C    Output File name
*  
*  
*  PRIVATE variables:
*  
*    Name          Type  Use
*  
*    m.pnFile       N    File pointer for output file
*    
*    m.nArea        N    Loop counter for work area from 1 to 225
*    m.nTgt         N    Loop counter for use with FoxPro function TARGET()
*    m.nTopLevel    N    Maximum level of nested relationships
*    m.nCurRec      N    The current record number (RECNO())
*    m.nCurrArea    N    The work area selected on calling this procedure
*    m.nWorkAreas   N    Total Number of work areas (VERSION() Specific)
*    m.nMaxNest	 N    Stores the maximum number of nested relations that
*  					  can be mapped. This is limited by the number of 
*  					  nested DO's the current version can support and
*    				      is that number minus 3. See manuals for details of
*  					  maximum nested DO's allowed. (Added 23 May 1994)
*    
*    m.cCrlf        C    Carrage Return and Line feed characters
*    m.cMap         C    Relation details and Diagram store.
*    m.cTalkSet     C    Setting of SET TALK on entry to this procedure
*    m.cCursName    C    Generated name (sys(2015) of created cursor
*    m.cCntStore    C    The Table name used for counting the total tables
*                        in level 1 for array na0LevTot
*    m.cStoreFrom   C    Value of FROMTABLE field and is used to determine
*                        if the current top level relation has already been
*                        reported on.
*    
*    m.lNotFound    L    Indicates if NO relations have been found (TRUE)
*                        or one or more were found (FALSE)
*    m.lDoneOnce    L    Indicates if top level relation has been reported
*                        on. Depends on value of m.cStoreFrom variable
*    
*    
*  PRIVATE arrays
*    
*    Name          Type  Use
*    
*    na0LevTot      N    Each element, refrenced by the current level number
*                        is a total for the number of relations that need 
*                        to be done at that level.
*    na0LevCnt      N    Each element, refrenced by the current level number
*                        is a counter for each time a relation of the
*                        respective level has been processed. It is used in
*                        conjunction with the na0LevTot array to determine if
*                        the relation beinf processed is the last relation
*                        for the current level, or not.
*    ca0LevPad      C    Each element, refenced by the current level number
*                        stores the text string that is to be used as the 
*                        padding for that level. Padding is the term used
*                        to describe the combination of spaces and vertical
*                        lines to the left of the table name:
*    
*    
*       e.g.      Table1
*                  Table2
*                     Table3
*                        Table4
*                     Table5
*                         Table6
*                  Table7
*    
*    The 'padding' is as follows (the  indicates where the padding
*    lines end.
*                 
*                  
*                     
*                        
*                     
*                         
*                  
*    
* ;
	
m.cTalkSet = set("talk")
set talk off
m.nCurrArea = select()

* Determine how many work areas there are to search.
do case
case "2.0" $ version()
	m.nWorkAreas = 25
	* set m.nMaxNest to 3 less that maximum DO nesting level because
	* these extra levels are required by the BUILDMAP procedure
	m.nMaxNest = 29
case "2.5" $ version() or "2.6" $ version()
	m.nWorkAreas = 225
	* set m.nMaxNest to 3 less that maximum DO nesting level because
	* these extra levels are required by the BUILDMAP procedure
	m.nMaxNest = 29
otherwise
	m.nWorkAreas = 0
	m.nMaxNest = 0
endcase

* create and populate the cursor with the information needed about 
* each individual relation.

m.cCursName = sys(2015)
create cursor (m.cCursName) ;
	(fromtable c(10), totable c(10), field c(14), order c(10))


m.lNotFound = .t.
for m.nArea = 1 to m.nWorkAreas	 && No loops if unknown FoxPro version
	if used(m.nArea) 
		m.nTgt = 1
		****************************************************************
		* The folling line was modified from :
		*
		*         do while target(m.nTgt, m.nArea) <> ""
		*
		* Due to the effect that SET EXACT had on the results of this 
		* comparrison.
		*
		* Simon Collingridge	23 May 1994
		****************************************************************
		do while !empty(target(m.nTgt, m.nArea))
			m.lNotFound = .f.
			append blank
			replace fromtable with alias(m.nArea), ;
				totable with target(m.nTgt, m.nArea), ;
				field with relation(m.nTgt, m.nArea), ;
				order with order(m.nArea)
			m.nTgt = m.nTgt + 1
		enddo while target(m.nTgt, m.nArea) <> ""
	endif used(m.nArea) 
endfor m.nArea = 1 to m.nWorkAreas

release m.nArea, m.nTgt

m.cCrlf		= chr(13) + chr(10)
if m.lNotFound
	*****************************************************************
	* Made more descriptive to help explain what the message given
	* is for.
	*
	* Simon Collingridge	23 May 1994
	*****************************************************************
	m.cMap= "ͻ" + m.cCrLf + ;
			"   DRAWMAP Relation Mapping Tool was " + m.cCrLf + ;
			"   Unable to map relations because   " + m.cCrLf	
	if m.nWorkAreas = 0
		m.cMap = m.cMap + "   the version of FoxPro is unknown. "
	else
		m.cMap = m.cMap + "   no relations were found to map.   "
	endif
	m.cMap = m.cMap + m.cCrLf + ;
			"ͼ" + m.cCrLf
else
	
	select (m.cCursName)

	index on totable tag xtToTable
	index on fromtable tag xtFromTabl

	m.cStoreFrom	= ""
	m.lDoneOnce	= .f.

	* Load m.cMap with header information
	m.cMap =  m.cCrlf + m.cCrlf + space(14) + "Table Relationships" + m.cCrlf + m.cCrlf + ;
		m.cCrlf+"Relation Field  Table Order  Diagram" + m.cCrlf + replicate("",15) + ;
		"" + replicate("",13) + "" + replicate("",30)

	* Now scan through the cursor, identifying each top level parent, then 
	* build the relational structures for all levels below them.

	scan		&& (m.cCursName)
		* initialise/reinitialise required variables and arrays
		dimension na0LevCnt(1), na0LevTot(2), ca0LevPad(1), ca0Parent(2)
		if empty(na0LevCnt(1)) or na0LevCnt(1) = -1
			na0LevCnt(1) = 0
		endif
		store "" to ca0LevPad, ca0Parent
		m.nTopLevel		= 0
		na0LevTot(1)	= 0
		m.nCurRec		= recno()
		m.cCntStore		= fromtable
		set order to tag xtToTable
		if !seek(fromtable)	&& Nothing relates INTO this table, ;
								must be top level
			set order to tag xtFromTabl
			count to na0LevTot(1) for fromtable = m.cCntStore	&& Total tables ;
															in level 1
			go m.nCurRec
			m.lDoneOnce = (m.cStoreFrom == fromtable)
			if !m.lDoneOnce	&& Insert blank lines to space top level groups
				m.cMap = m.cMap + m.cCrlf + space(15) + "" + space(13) + " " + ;
					m.cCrlf + space(15) + "" + space(13) + " "
			endif
			m.cStoreFrom = fromtable
			do buildmap with 1, m.lDoneOnce	&& Enter recussive function to map ;
											all sub-level relations
		endif
		set order to tag xtFromTabl	&& Re-order to ensure the SCAN does so ;
										in the correct order
		go m.nCurRec
	endscan (m.cCursName)
	
	***********************************************************************
	* Now output the relation map into a file. It is at this point that the
	* output of the function can be directed to wherever it may be required
	* by storing the contents of the variable m.cMap elsewhere.
	***********************************************************************

	
endif m.lNotFound

m.cMapText = substr(sys(3), 4, 4) + substr(sys(2015),7,4) + ".map"
m.pnFile = fcreate(m.cMapText)
=fputs(m.pnFile,m.cMap)
=fclose(m.pnFile)

if used(m.cCursName)		&& Close cursor
	select (m.cCursName)
	use
endif

* Finally reset the environment
select (m.nCurrArea)
if m.cTalkSet = "ON"
	set talk on
endif

return .t.



procedure buildmap

* ͸
*  
*  Procedure:      BUILDMAP
*  
*  Author:         Simon Collingridge
*  Date:           10/5/94
*  
*  DESCRIPTION:
*  
*      The first thing to note about this procedure is that it is
*  recursive, (i.e. It makes calls on itself). Although not very good
*  practice for the purpose of this procedure it is unavoidable as
*  their could be any number of relation levels to process.
*  
*      The general way the procedure works is like this. First the level
*  counter (na0LevCnt), total (na0LevTot) and padding (ca0LevPad) are all defined
*  checking first if a new element needs to be added and initialised 
*  because the current level has not been processed before. This definition
*  is including the end to the padding which is stored in m.cPadEnd. The 
*  padding and padding end are then stored in the m.cMap variable, (if the 
*  MISSIT parameter is not set to true). Once this is done, the procedure 
*  then check to see if there is a sub-level off the current level. If
*  there is then the following occurs:
*  
*      The total number of tables in the sub-level are counted and
*      stored to the sub-levels element of the na0LevTot array.
*  
*      The function recurses passing the next level number and whether
*      the level has be processed before (ONELOOP)
*  
*  
*  If there are no sub-levels, then the current end level is processed
*  and mapped.
*  
*  Control is then passed back up the chain, either to the previous
*  instance of this procedure, or back to DRAWRELS if the current
*  level is level 1.
*  
*  If control is passed back to a previous instance of this procedure,
*  if the level count for the next but one sub-level exists, it is 
*  initialised ready for any other table relations at that level. A 
*  check is made to see if any more tables at the current level exist
*  and if so, the loop is repeated, recusing to check for further
*  sub-levels.
*  
*  Inputs: 
*  
*  	Parameters:
* 
*          m.nLevel    -    Numeric       -  The level to process
*          m.lMissIt   -    Logical       -  If TRUE, don't map the
*   	                                      relations details. Used to
*      	                                  trap for successive calls
*          	                              of this proc. at the same
*              	                          level.
*  
*      External Variables:
*  
*         na0LevCnt    -   Numeric Array  - Level process counters
*         na0LevTot    -   Numeric Array  - Level Table totals
*         ca0LevPad    -   Char. Array    - Level padding strings
*         m.nTopLevel  -   Numeric        - Lowest level number so far
*         m.cMap       -   Character      - Relationship Map and details
*         m.nMaxNest	-	Numeric		   - Maximum number of levels 
*         									 that can be mapped
*         
*         
*  Outputs:
*  
*      Text file of the relations diagram. The file name is stored in 
*      PUBLIC VARIABLE m.cMapText. Output can be easily re-directed by 
*      changing the destination of the memory variable m.cMap to wherever
*      it is needed.
*  
*  
*  Called By:    DRAWMAP
*  
*  Calls:        GETPADDING
*                GETFIELD
*  
*  Tables Used:  The (m.cCursName) Cursor
*  
*  Calling
*     Sequence:  DO buildmap with <m.nLevel>, <m.lMissIt>
*  	
*  	
*  MODIFICATIONS:
*  	
*  	
*  	23 May 1994		Simon Collingridge
*  	
*  	The following modifications have been made:
*  	
*  	1.	Code was added to the to check the level of nested relations
*  		being mapped do not exceed the value stored by the m.nMaxNest 
*  		variable. The value of m.nMaxNest is determined by the 
*  		VERSION of FoxPro being used and is the number of allowable 
*  		nested DO's minus 3.
*  
* ;

parameters m.nLevel, m.lMissIt

private m.nSaveRec, m.cToStore, m.lOneLoop, m.cPadEnd

* ͸
*                  MEMORY VARIABLE DEFINITIONS AND USE
*  
*  PRIVATE variables:
*  
*    Name        Type  Use
*  
*    m.nSaveRec   N    Current Record Numer of the CURSNAME Cursor
*    
*    m.cToStore   C    Value of the TOTABLE field in the Cursor
*    m.cPadEnd    C    Padding end to use in mapping the table name
*    
*    m.lOneLoop   L    Indictae if current level has been processed one
*                      or more times (TRUE) or not at all (FALSE)
*    
*    
* ;
	
m.lOneLoop	= .f.
m.nSaveRec	= recno()
m.cToStore	= totable

if m.nLevel > alen(na0LevCnt)	&& If first time level is processed, add new ;
							elements to the arrays and initialise them
	dimension na0LevCnt(m.nLevel), na0LevTot(m.nLevel+1), ;
		ca0LevPad(m.nLevel), ca0Parent(m.nLevel+1)
	store 1 to na0LevCnt(m.nLevel)
	na0LevTot(m.nLevel+1) = 0
	store "" to ca0LevPad(m.nLevel)
else					&& If level previously processed, increment counter
	na0LevCnt(m.nLevel) = na0LevCnt(m.nLevel) + 1
	if na0LevCnt(m.nLevel) = na0LevTot(m.nLevel)	&& Indicate last one done by setting
		na0LevCnt(m.nLevel) = -1				&& counter to -1
	endif
endif

if m.nLevel > m.nTopLevel
	m.nTopLevel = m.nLevel
endif


*************************************************************************
* The following IF - ENDIF was added to check that the number of nested
* relations being mapped does not exceed the version allowable number
* of nested DO's. Note the value of m.nMaxNest is the number of allowable
* nested DO's minus 3
*
* Simon Collingridge	23 May 1994
*************************************************************************
if m.nLevel >= m.nMaxNest
	m.cMap = m.cMap + m.cCrlf + padr("==*==*==",15) + "" + ;
			padr(" ==*==*==",13) + " " + getpadding(m.nLevel) + ;
			"Ķ Maximum Mapping Level Exceeded"
	return
endif

if !m.lMissIt			&& store mapping information
	if m.nLevel = 1
		m.cMap = m.cMap + m.cCrlf + space(15) + " " + order + "   " + fromtable
		ca0LevPad(m.nLevel) = ""
	else
		* Using the level counter and total, check if this is the last
		* Table to process at this level, and if so use the correct
		* padding end. Also store what padding should be used by sub-levels
		* when padding out the space created by this level.
		if na0LevTot(m.nLevel-1) > 1 and between(na0LevCnt(m.nLevel-1),0,na0LevTot(m.nLevel-1)-1)
			m.cPadEnd	= "" + "" + chr(26)
			ca0LevPad(m.nLevel) = "  "
		else
			m.cPadEnd	= "" + "" + chr(26)
			ca0LevPad(m.nLevel) = "   "
		endif
		m.cMap = m.cMap + m.cCrlf + getfield(fromtable, ca0Parent(m.nLevel)) + "  " + order + ;
			"   " + getpadding(m.nLevel) + m.cPadEnd + fromtable
	endif
endif

ca0Parent(m.nLevel+1) = fromtable	&& Store name to next level Parent element

locate for fromtable = m.cToStore  && Check for any sub-levels
if found()
	do while found()		&& Count how many Tables are in the sub-level
		m.nSaveRec = recno()
		count to na0LevTot(m.nLevel+1) for fromtable = m.cToStore
		go m.nSaveRec
		do buildmap with m.nLevel + 1, m.lOneLoop	&& Recurse procedure ;
													with new level number
		if m.nLevel+2 <= m.nTopLevel			&& Initialise counter for ;
													2nd sub-level
			store 0 to na0LevCnt(m.nLevel+2)
		endif
		go m.nSaveRec
		skip 1
		locate rest for fromtable = m.cToStore	&& Check for further Tables ;
												at this level (CONTINUE ;
												doesn't work)
		m.lOneLoop = .t.		&& Indicate the level has been processed at ;
								least once
	enddo
else
	* If there are no more sub-levels off this level, then process this
	* levels current table
	go m.nSaveRec
	if na0LevCnt(m.nLevel) = 0
		na0LevCnt(m.nLevel) = na0LevCnt(m.nLevel) + 1
	endif
	if na0LevTot(m.nLevel) > 1 and between(na0LevCnt(m.nLevel),0,na0LevTot(m.nLevel)-1)
		m.cPadEnd	= "" + "" + chr(26)
	else
		m.cPadEnd	= "" + "" + chr(26)
	endif
	m.cMap = m.cMap + m.cCrlf + getfield(padr(totable,10), ;
		ca0Parent(m.nLevel+1)) + "  " + ;
		padr(order(trim(totable)),10) + "   " + ;
		getpadding(m.nLevel+1) + m.cPadEnd + totable
endif



function getpadding

* ͸
*  
*  Function:       GETPADDING
*  
*  Author:         Simon Collingridge
*  Date:           10/5/94
*  
*  DESCRIPTION:
*  
*      This function creates the padding string needed by the current
*  level when mapping a table. Id concatenates all the previous levels
*  individual padding strings to return one complete padding string.
*  
*  Inputs: 
*  
*  	Parameters:
* 
*          m.nLevel    -    Numeric    -    The level to process
*  
*      External Variables:
*  
*         ca0LevPad    -   Char. Array -    Level padding strings
*         
*  Outputs:
*  
*      Returns:
*  
*          m.cPadding  -   Character   -    Complete padding string
*  
*  Called By:    BUILDMAP
*  
*  Calls:        None
*  
*  Tables Used:  None
*  
*  Calling
*     Sequence:  =getpadding(<m.nLevel>)
*  
* ;

parameters m.nLevel

private m.cPadding
m.cPadding = ""
for cnt = 1 to m.nLevel-1
	m.cPadding = m.cPadding + ca0LevPad(cnt)
endfor
return m.cPadding



function getfield

* ͸
*  
*  Function:       GETFIELD
*  
*  Author:         Simon Collingridge
*  Date:           10/5/94
*  
*  DESCRIPTION:
*  
*      This function finds the table that is parent over the child table
*  whose name is passed as a parameter, then it gets the field name of
*  the field that is used to relate into the child and returns it. This
*  function ensures that if a table is parent over more than one child, 
*  the correct field name is mapped against the correct child.
*  
*  Inputs: 
*  
*  	Parameters:
* 
*          m.cChild    -   Character      - Name of child Table
*  
*  
*         
*  Outputs:
*  
*      Returns:
*  
*          m.cField    -   Character      - Name of field use by parent
*  
*  Called By:    BUILDMAP
*  
*  Calls:        None
*  
*  Tables Used:  The (m.cCursName) Cursor
*  
*  Calling
*     Sequence:  =getfield(<m.cChild>)
*  
* ;

parameter m.cChild, m.cParent

private m.nCurRec, m.lFound, m.cField

m.nCurRec	= recno()

locate for m.cParent == fromtable and m.cChild == padr(totable,10)
if found()
	m.cField = field
else
	m.cField = space(14)
endif

go m.nCurRec
return m.cField
